@@ -1,22 +1,38 @@
 #!/usr/bin/env python3
 # coding: utf-8
 
+from __future__ import annotations
+
+import textwrap
+import dataclasses
 import datetime
 import decimal
+import functools
 import mimetypes
 import re
+from functools import cached_property
+from typing import Callable
+from typing import Union
 
 import flask
 import flask.views
+from flask import Flask
 from volkanic.utils import merge_dicts
+# noinspection PyPackageRequirements
+from werkzeug.routing import Rule
 
 
-def infer_mime_type(filename: str, default="text/plain") -> str:
-    if filename.startswith('.'):
+def infer_mimetype(filename: str, default="text/plain") -> str:
+    if '.' not in filename:
+        filename = '_.' + filename
+    elif filename.startswith('.'):
         filename = '_' + filename
     return mimetypes.guess_type(filename)[0] or default
 
 
+infer_mime_type = infer_mimetype
+
+
 def respond(*args, **kwargs):
     resp = kwargs
     for arg in args:
@@ -39,18 +55,37 @@ def respond(*args, **kwargs):
     return resp
 
 
-def respond_plain_text(text: str):
-    resp = flask.make_response(text, 200)
-    resp.mimetype = "text/plain"
+def respond_content(
+        content: Union[bytes, str],
+        mimetype: str = None,
+        headers: dict = None):
+    resp = flask.make_response(content, 200)
+    if headers:
+        for key, val in headers.items():
+            resp.headers.set(key, val)
+    if mimetype:
+        resp.mimetype = mimetype
+    elif isinstance(content, str):
+        resp.mimetype = "text/plain"
+    else:
+        resp.mimetype = 'application/octet-stream'
     return resp
 
 
-def respond_with_pre_gzipped(content: bytes, content_type=None):
+def respond_plain_text(text: str, headers: dict = None):
+    return respond_content(text, "text/plain", headers)
+
+
+def respond_with_pre_gzipped(
+        content: bytes, content_type=None, headers: dict = None):
     content_type = content_type or 'text/plain'
-    resp = flask.make_response(content)
-    resp.headers.set('Content-Type', content_type)
-    resp.headers['Content-Encoding'] = 'gzip'
-    return resp
+    headers = headers or {}
+    headers['Content-Encoding'] = 'gzip'
+    return respond_content(content, content_type, headers)
+
+
+def respond_binary(content: bytes, content_type: str, headers: dict = None):
+    return respond_content(content, content_type, headers)
 
 
 def respond_xaccel_redirect(path: str, filename: str = None):
@@ -106,43 +141,61 @@ def jsonp(data, callback):
     )
 
 
-_datetime_fmt = '%Y-%m-%d %H:%M:%S'
+def fmt_datetime(o: datetime.datetime):
+    if isinstance(o, datetime.datetime):
+        return o.strftime('%Y-%m-%d %H:%M:%S')
+    raise TypeError
 
 
-def _json_default(o):
+def json_default_strict(o):
     if isinstance(o, decimal.Decimal):
         return float(o)
     if isinstance(o, datetime.timedelta):
         return o.total_seconds()
-    if isinstance(o, datetime.date) and not isinstance(o, datetime.datetime):
+    if isinstance(o, datetime.datetime):
+        return o.strftime('%Y-%m-%d %H:%M:%S')
+    if isinstance(o, datetime.date):
         return o.isoformat()
     if hasattr(o, 'as_json_serializable'):
         return o.as_json_serializable()
     raise TypeError
 
 
+def call_json_default_functions(o, funcs: Union[tuple, list] = None):
+    funcs = funcs or []
+    for func in funcs:
+        try:
+            return func(o)
+        except TypeError:
+            continue
+    raise TypeError
+
+
 def json_default(o):
     """usage: json.dumps(some_o, default=json_default)"""
-    try:
-        return _json_default(o)
-    except TypeError:
-        pass
-    if isinstance(o, datetime.datetime):
-        return o.strftime(_datetime_fmt)
-    return str(o)
+    funcs = [json_default_strict, str]
+    return call_json_default_functions(o, funcs)
+
+
+def chain_json_default_functions(*funcs):
+    return functools.partial(call_json_default_functions, funcs=funcs)
 
 
 class JSONEncoderPlus(flask.json.JSONEncoder):
-    datetime_fmt = _datetime_fmt
+    json_default_funcs = [json_default_strict, str]
 
     def default(self, o):
         try:
-            return _json_default(o)
+            return call_json_default_functions(o, self.json_default_funcs)
         except TypeError:
-            pass
-        if isinstance(o, datetime.datetime):
-            return o.strftime(self.datetime_fmt)
-        return super().default(o)
+            return super().default(o)
+
+
+def get_json_encoder(*json_default_funcs):
+    # bases should be a tuple
+    bases = JSONEncoderPlus,
+    attrs = {'json_default_funcs': list(json_default_funcs)}
+    return type('JSONEncoderExtended', bases, attrs)
 
 
 def serialize_current_session(app: flask.Flask = None):
@@ -186,3 +239,39 @@ def decorate_all_view_funcs(app, decorator):
 def url_for_this(**values):
     # https://github.com/pallets/flask/issues/2111
     return flask.url_for(flask.request.endpoint, **values)
+
+
+@dataclasses.dataclass
+class ViewEntry:
+    rule: Rule
+    func: Callable
+
+    @cached_property
+    def methods(self) -> list[str]:
+        vals = list(self.rule.methods)
+        vals.sort()
+        return vals
+
+    @cached_property
+    def explicit_methods(self) -> list[str]:
+        implicit = {'HEAD', 'OPTIONS'}
+        return [s for s in self.methods if s not in implicit]
+
+    def fmt_methods(self, sep='|', implicit=False):
+        methods = self.methods if implicit else self.explicit_methods
+        return sep.join(methods)
+
+    @property
+    def help(self) -> dict:
+        return getattr(self.func, 'help', {})
+
+    @property
+    def docstring(self) -> str:
+        return textwrap.dedent(self.func.__doc__ or '')
+
+    @classmethod
+    def get_all(cls, app: Flask) -> dict[str, ViewEntry]:
+        # entries = {}
+        rules = {r.endpoint: r for r in app.url_map.iter_rules()}
+        funcs = app.view_functions
+        return {k: cls(rules[k], funcs[k]) for k in rules}
